

import requests

for sentence in d:
    r = requests.post("http://84.86.153.191/predict",
                      json={"sentence": sentence, "model_name": "tense"})
    print("-" * 50)
    print(sentence)
    for sentence in r.json():
        print(sentence['intent'], sentence['confidence'])


"last X years",
"X years ago",
"X years prior",
"X years earlier"



# Stackoverflow CC License (https://creativecommons.org/licenses/by-sa/3.0/)
# Author: D Greenberg (http://stackoverflow.com/users/5133085/d-greenberg)
# Link: http://stackoverflow.com/a/31505798/1575066
import re
caps = "([A-Z])"
prefixes = "(Mr|St|Mrs|Ms|Dr)[.]"
suffixes = "(Inc|Ltd|Jr|Sr|Co)"
starters = "(Mr|Mrs|Ms|Dr|He\s|She\s|It\s|They\s|Their\s|Our\s|We\s|But\s|However\s|That\s|This\s|Wherever)"
acronyms = "([A-Z][.][A-Z][.](?:[A-Z][.])?)"
websites = "[.]([a-zA-Z]{2,3})"


def simple_sentencer(text):
    text = " " + text + "  "
    text = text.replace("\n", " ")
    text = re.sub(prefixes, "\\1<prd>", text)
    text = re.sub(websites, "<prd>\\1", text)
    if "Ph.D" in text:
        text = text.replace("Ph.D.", "Ph<prd>D<prd>")
    text = re.sub("\s" + caps + "[.] ", " \\1<prd> ", text)
    text = re.sub(acronyms + " " + starters, "\\1<stop> \\2", text)
    text = re.sub(caps + "[.]" + caps + "[.]" + caps + "[.]", "\\1<prd>\\2<prd>\\3<prd>", text)
    text = re.sub(caps + "[.]" + caps + "[.]", "\\1<prd>\\2<prd>", text)
    text = re.sub(" " + suffixes + "[.] " + starters, " \\1<stop> \\2", text)
    text = re.sub(" " + suffixes + "[.]", " \\1<prd>", text)
    text = re.sub(" " + caps + "[.]", " \\1<prd>", text)
    if "”" in text:
        text = text.replace(".”", "”.")
    if "\"" in text:
        text = text.replace(".\"", "\".")
    if "!" in text:
        text = text.replace("!\"", "\"!")
    if "?" in text:
        text = text.replace("?\"", "\"?")
    text = text.replace(".", ".<stop>")
    text = text.replace("?", "?<stop>")
    text = text.replace("!", "!<stop>")
    text = text.replace("<prd>", ".")
    sentences = text.split("<stop>")
    sentences = sentences[:-1]
    sentences = [s.strip() for s in sentences]
    return sentences



from metadate import parse_date
import pandas as pd
qs = pd.read_csv("~/cancellation_data.csv")["reply_message"]
qs = [x for x in qs if isinstance(x, str)]
for q in qs:
    try:
        mp = parse_date(q)
    except:
        print(q)
        raise
    if mp is not None:
        print(mp.matches)



import re
from enum import IntEnum

from dateutil.relativedelta import relativedelta


class Units(IntEnum):
    YEAR = 9
    SEASON = 8
    QUARTER = 7
    MONTH = 6
    WEEK = 5
    DAY = 4
    HOUR = 3
    MINUTE = 2
    SECOND = 1
    MICROSECOND = 0

RD_ARG_TO_UNIT = {
    "years": Units.YEAR,
    "year": Units.YEAR,
    "months": Units.MONTH,
    "month": Units.MONTH,
    "weeks": Units.WEEK,
    "week": Units.WEEK,
    "weekday": Units.DAY,
    "days": Units.DAY,
    "day": Units.DAY,
    "hours": Units.HOUR,
    "hour": Units.HOUR,
    "minutes": Units.MINUTE,
    "minute": Units.MINUTE,
    "seconds": Units.SECOND,
    "second": Units.SECOND,
    "microseconds": Units.MICROSECOND,
    "microsecond": Units.MICROSECOND,
}

freqs = ["yearly", "monthly", "weekly", "daily", "hourly", "minutely", "secondly"]
FREQ = {x: num for num, x in enumerate(freqs)}


def log(tag, x, verbose=False):
    if verbose:
        print("--- {} ------------".format(tag))
        print(x)


def add_tag(sentence, matches, color="mediumspringgreen"):
    # given textual matches between ranges like [(5, 10), (10, 15)]
    # this will clean up
    # first is 0,5
    news = ''
    lbound = 0
    hit = "<span style='background-color:{}'>{}</span>"
    for m in matches:
        news += sentence[lbound:m[0]]
        news += hit.format(color, sentence[m[0]:m[1]])
        lbound = m[1]
    news += sentence[matches[-1][1]:]
    return news


def strip_pm(txt, numbers_dict=None):
    txt = txt.lower()
    hoffset = 12 * ('pm' in txt or "afternoon" in txt or 'p.m' in txt)
    txt = txt.replace("afternoon", "")
    txt = txt.replace("'", " ")
    txt = re.sub(".?o.?clock", "", txt)
    if numbers_dict is None:
        parts = re.sub("[:hapm.]+", " ", txt).split()
    else:
        parts = re.sub(" ?[ap][.]?m[.]?$", " ", txt).split()
    parts = [x for x in parts if x]
    microsecond = None
    second = None
    minute = None
    if len(parts) == 4:
        hour, minute, second, microsecond = parts
    elif len(parts) == 3:
        hour, minute, second = parts
    elif len(parts) == 2:
        hour, minute = parts
    else:
        hour = parts[0]
    if numbers_dict is not None:
        hour = numbers_dict[hour]
    if int(hour) == 12 and hoffset:
        hoffset = 0
    hour = (hoffset + int(hour)) % 24
    hour = 0 if hour is None else int(hour)
    minute = 0 if minute is None else int(minute)
    second = 0 if second is None else int(second)
    microsecond = 0 if microsecond is None else int(microsecond)
    return hour, minute, second, microsecond


def erase_level(d, level):
    if level == Units.YEAR:
        d = d.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
    elif level == Units.SEASON:
        d = d.replace(day=21, hour=0, minute=0, second=0, microsecond=0)
    elif level == Units.QUARTER:
        d = d.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    elif level == Units.MONTH:
        d = d.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    elif level == Units.WEEK:
        d = d.replace(hour=0, minute=0, second=0, microsecond=0)
    elif level == Units.DAY:
        d = d.replace(hour=0, minute=0, second=0, microsecond=0)
    elif level == Units.HOUR:
        d = d.replace(minute=0, second=0, microsecond=0)
    elif level == Units.MINUTE:
        d = d.replace(second=0, microsecond=0)
    elif level == Units.SECOND:
        d = d.replace(microsecond=0)
    return d


def resolve_end_period(start_date, level, past_boundary, now):
    date_changed = False
    start_date = erase_level(start_date, level)
    if level == Units.YEAR:  # year
        end_date = start_date + relativedelta(years=1)
    elif level == Units.SEASON:  # season
        end_date = start_date + relativedelta(months=3)
    elif level == Units.QUARTER:  # quarter
        for cap in [1, 4, 7, 10]:
            if start_date.month <= cap:
                break
        end_month = (cap + 3)
        if end_month > 12:
            end_date = start_date + relativedelta(month=end_month % 12, day=1, years=1)
        else:
            end_date = start_date + relativedelta(month=end_month, day=1)
    elif level == Units.MONTH:
        end_date = start_date + relativedelta(months=1)
    elif level == Units.WEEK:
        end_date = start_date + relativedelta(days=7)
    elif level == Units.DAY:
        end_date = start_date + relativedelta(days=1)
    elif level == Units.HOUR:
        end_date = start_date + relativedelta(hours=1)
    elif level == Units.MINUTE:
        end_date = start_date + relativedelta(minutes=1)
    elif level == Units.SECOND:
        end_date = start_date + relativedelta(seconds=1)
    elif level == Units.MICROSECOND:
        end_date = start_date + relativedelta(microseconds=1)
    # has to be tighter
    if end_date < past_boundary:
        start_date = start_date.replace(year=start_date.year + 1)
        end_date = end_date.replace(year=end_date.year + 1)
        date_changed = True
    elif start_date < past_boundary:
        start_date = now
        date_changed = True
    return start_date, end_date, date_changed

BOUNDARIES = {
    1: 31,
    2: 29,
    3: 31,
    4: 30,
    5: 31,
    6: 30,
    7: 31,
    8: 31,
    9: 30,
    10: 31,
    11: 30,
    12: 31
}



from datetime import datetime
from metadate.utils import Units
from metadate.utils import FREQ
from metadate.utils import RD_ARG_TO_UNIT
from dateutil.relativedelta import relativedelta


class Meta(object):

    def __init__(self, x, span):
        self.x = x
        self.span = span

    def __repr__(self):
        msg = "{}(x={})"
        return msg.format(self.__class__.__name__, self.x)


class MetaPeriod(Meta):

    def __init__(self, start_date, end_date, level, spans, locale, text):
        self.start_date = start_date
        self.end_date = end_date
        self.level = level
        self.spans = spans
        self.locale = locale
        self.text = text
        self.matches = [text[i:j] for i, j in spans]

    def __repr__(self):
        name = self.__class__.__name__
        cases = ['start_date', 'end_date', 'level', 'matches', 'locale']
        n = len(name) + 1
        atts = ", ".join([str(getattr(self, x))
                          for x in cases if getattr(self, x, None) is not None])
        return "{}({})".format(name, atts)

    def to_dict(self):
        result = {
            'start_date': self.start_date.isoformat(),
            'end_date': self.end_date.isoformat(),
            'level': self.level,
            'spans': [{"begin": x[0], "end": x[1]} for x in self.spans],
            'matches': self.matches,
        }
        return result


class MetaUnit(Meta):

    def __init__(self, unit, span, modifier=1):
        self.unit = unit
        self.span = span
        self.level = Units[unit]
        self.modifier = modifier

    def __repr__(self):
        cases = ["unit", "modifier"]
        atts = {x: str(getattr(self, x)) for x in cases if getattr(self, x, None) is not None}
        values = ', '.join(["{}={}".format(x, atts[x]) for x in cases if x in atts])
        return "{}({})".format(self.__class__.__name__, values)


class MetaOrdinal(Meta):

    def __init__(self, amount, span):
        self.amount = float(amount)
        self.span = span

    def __repr__(self):
        cases = ["amount"]
        atts = {x: str(getattr(self, x)) for x in cases if getattr(self, x, None) is not None}
        values = ', '.join(["{}={}".format(x, atts[x]) for x in cases if x in atts])
        return "{}({})".format(self.__class__.__name__, values)


class MetaRelative(Meta):

    def __init__(self, span, level=None, rd=None, **rd_args):
        if level is None:
            level = min([RD_ARG_TO_UNIT[x] for x in rd_args])
        self.rd_args = rd_args
        self.level = level
        if rd is not None:
            self.rd = rd_args
        else:
            if "seasons" in rd_args:
                del rd_args["seasons"]
            self.rd = relativedelta(**rd_args)
        self.span = span

    def __add__(self, other):
        # if overlap, crash
        rd = self.rd + other.rd
        level = self.level if self.level > other.level else other.level
        span = sorted(self.span + other.span)
        return MetaRelative(level, span, rd=rd)

    def __repr__(self):
        return "{}(rd={}, level={})".format(self.__class__.__name__, self.rd, self.level)


class MetaModifier(Meta):

    def __init__(self, x, value, span):
        self.x = x
        self.value = value
        self.span = span

    def __repr__(self):
        msg = "{}(x={}, value={})"
        return msg.format(self.__class__.__name__, self.x, self.value)


class MetaRange(Meta):
    pass


class MetaBetween(Meta):

    def __init__(self, start, end, level, span):
        self.start = start
        self.end = end
        self.level = level
        self.span = span

    def __repr__(self):
        msg = "{}(start={}, end={}, level={})"
        return msg.format(self.__class__.__name__, self.start, self.end, self.level)


class MetaEvery(Meta):

    def __init__(self, freq, interval, level, span):
        self.span = span
        self.freq = freq
        self.interval = interval
        self.level = level

    def __repr__(self):
        msg = "{}(freq={}, interval={}, level={})"
        return msg.format(self.__class__.__name__, self.freq, self.interval, self.level)


class MetaAnd(Meta):
    pass


class MetaDuration(Meta):

    def __init__(self, rd, level, span):
        self.span = span
        self.rd = rd
        self.level = level

    def __repr__(self):
        msg = "{}(rd={}, level={})"
        return msg.format(self.__class__.__name__, self.rd, self.level)



# optional flask dependency
from flask import Flask
from flask import request
from flask import jsonify
from metadate import parse_date

app = Flask(__name__)


@app.route("/metadate", methods=["GET", "POST"])
def root():
    text = request.args.get("text")
    print(text)
    future = request.args.get("future")
    future = future.lower() == 'true' if future else True
    res = parse_date(text, future)
    if res:
        res = res.to_dict()
    return jsonify(res)

if __name__ == "__main__":
    app.run(host="0.0.0.0", port=8002)



NAME = "en"

# make it not based on index, but key val, and include "hrs", "wks"
UNITS = {
    'microsecond': 'MICROSECOND',
    'microseconds': 'MICROSECOND',
    'second': 'SECOND',
    'seconds': 'SECOND',
    'minute': 'MINUTE',
    'minutes': 'MINUTE',
    'mins': 'MINUTE',
    'min': 'MINUTE',
    'hour': 'HOUR',
    'hours': 'HOUR',
    "hrs": 'HOUR',
    "hr": 'HOUR',
    'day': 'DAY',
    'days': 'DAY',
    'week': 'WEEK',
    'weeks': 'WEEK',
    'wks': "WEEK",
    'quarter': 'QUARTER',
    'quarters': 'QUARTER',
    'month': 'MONTH',
    'months': 'MONTH',
    'season': 'SEASON',
    'seasons': 'SEASON',
    'year': 'YEAR',
    'years': 'YEAR',
    'yrs': "YEAR"
}


MODIFIERS = {
    "in": 1,
    "on": 0,
    "this": 0,
    "next":  1,
    "coming": 1,
    "after": 1,
    "before": -1,
    'recent': -1,
    'past': -1,
    # untested from here
    "previous": -1,
    "last": -1,
    "first": 1,
    "ago": -1,
    "earlier": -1,
    "prior": -1
}

ORDINAL_NUMBERS = {
    'zero': 0,
    'one': 1,
    'two': 2,
    'three': 3,
    'few': 3,
    'a couple': 3,
    'four': 4,
    'five': 5,
    'six': 6,
    'seven': 7,
    'eight': 8,
    'nine': 9,
    'ten': 10,
    'eleven': 11,
    'twelve': 12,
    'thirteen': 13,
    'fourteen': 14,
    'fifteen': 15,
    'sixteen': 16,
    'seventeen': 17,
    'eighteen': 18,
    'nineteen': 19,
    'twenty': 20,
    'thirty': 30,
    'fourty': 40,
    'fifty': 50,
    'sixty': 60,
    'seventy': 70,
    'eighty': 80,
    'ninety': 90,
    'hundred': 100,
    'thousand': 1000,
}

# for?
IN_THE = ['in the', 'over the', 'during the', 'within the']  # 'for'

MONTHS = {
    "january": 1, "february": 2, "march": 3, "april": 4, "may": 5, "june": 6, "july": 7,
    "august": 8, "september": 9, "october": 10, "november": 11, "december": 12
}
MONTHS_SHORTS = {k[:3]: v for k, v in MONTHS.items()}
MONTHS_SHORTS["febr"] = 2
MONTHS_SHORTS["sept"] = 9

WEEKDAY = {
    "monday": 0,
    "tuesday": 1,
    "wednesday": 2,
    "thursday": 3,
    "friday": 4,
    "saturday": 5,
    "sunday": 6
}

WEEKDAY_SHORTS = {k[:3]: v for k, v in WEEKDAY.items()}

SEASONS = {
    "winter": 12,
    "spring": 3,
    "summer": 6,
    "fall": 9,
    "autumn": 9
}

QUARTERS = {
    "q1": 1,
    "first quarter": 1,
    "q2": 4,
    "second quarter": 4,
    "q3": 7,
    "third quarter": 7,
    "q4": 10,
    "fourth quarter": 10,
    "final quarter": 10,
    "last quarter": 10,
}


NOW = [
    'now'
]

TODAY_TOMORROW = {
    "today": 0,
    "tomorrow": 1,
    "day after tomorrow": 2,
    "yesterday": -1,
    "day before yesterday": -2
}

NOON = {
    "noon": 12,
    "midnight": 0,
    "end of noon": 17,
    "end of the afternoon": 17,
    "end of afternoon": 17,
    # "morning": 6 - 12?,
    # "night": 0 - 6?,
}

AND = ["and", ","]

RANK_NAMING = ['st', 'nd', 'rd', 'th']
ON_THE = ['on the']
AT = ['at']

# still to implement: night, morning
WHITELIST = ["at", "of", ",", "the", "night", "morning"]

BETWEEN = ["between"]
EVERY = ["every"]
DURATION = ["for", "until"]



from setuptools import find_packages
from setuptools import setup

MAJOR_VERSION = '0'
MINOR_VERSION = '2'
MICRO_VERSION = '60'
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)

setup(name='metadate',
      version=VERSION,
      description="All your date are belong to us",
      author='Pascal van Kooten',
      url='https://github.com/kootenpv/metadate',
      author_email='kootenpv@gmail.com',
      install_requires=[
          'python-dateutil', 'enum34'
      ],
      entry_points={
          'console_scripts': ['metadate = metadate.__main__:main']
      },
      classifiers=[
          'Intended Audience :: Developers',
          'Intended Audience :: Customer Service',
          'Intended Audience :: System Administrators',
          'Operating System :: Microsoft',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Unix',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Topic :: Software Development',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Software Distribution',
          'Topic :: System :: Systems Administration',
          'Topic :: Utilities'
      ],
      license='MIT',
      packages=find_packages(),
      zip_safe=False,
      platforms='any')



import re
from datetime import datetime

from dateutil.relativedelta import relativedelta as rd

import metadate.locales.en as locale_en
from metadate.classes import (MetaBetween, MetaModifier, MetaOrdinal,
                              MetaPeriod, MetaRange, MetaRelative, MetaUnit)
from metadate.scanner import Scanner
from metadate.utils import Units, BOUNDARIES, erase_level, log, resolve_end_period
from metadate.classes import MetaAnd


def get_relevant_parts(matches):
    strike = 0
    bundles = [[]]
    for m in matches:
        if isinstance(m, str):
            strike = 1
        else:
            if strike:
                bundles.append([m])
            else:
                bundles[-1].append(m)
            strike = 0
    return bundles


def between_allowed(x, y, text):
    start = x.span[1]
    end = y.span[0]
    return re.match("^[ -]*(and)?[ -]*$", text[start:end])


def merge_ordinal_unit(matches, text):
    news = []
    t = 0
    last = False
    n = len(matches)
    spans = []
    for i, m in enumerate(matches):
        if i != n - 1 and isinstance(m, MetaOrdinal):
            if last and not between_allowed(last, m, text):
                t = 0
            t += float(m.amount)
            last = m
            spans.append(m.span)
            continue
        elif isinstance(m, MetaUnit):
            m.modifier *= t if t else 1
            m.span = [m.span] + spans
            spans = []
        news.append(m)
    return news


def cleanup_relevant_parts(bundles, locale):
    cleaned_bundles = []
    for bundle in bundles:
        modifier = False
        meta_units = []
        new = []
        for x in bundle:
            if isinstance(x, MetaModifier):  # "this"
                if meta_units:
                    for mu in meta_units:
                        unit = mu.unit.lower() + 's'
                        rd_kwargs = {unit: mu.modifier * locale.MODIFIERS[x.x]}
                        new.append(MetaRelative(level=Units[mu.unit], span=[
                                   x.span, mu.span], **rd_kwargs))
                modifier = x
                meta_units = []
            elif isinstance(x, MetaAnd):
                continue
            elif modifier:
                if isinstance(x, MetaUnit):
                    if x.unit == "quarter":
                        rd_kwargs = {"months": 3 * x.modifier * locale.MODIFIERS[modifier.x],
                                     "day": 1}
                    else:
                        unit = x.unit.lower() + 's'
                        rd_kwargs = {unit: x.modifier * locale.MODIFIERS[modifier.x]}
                    new.append(MetaRelative(level=Units[x.unit], span=[
                               x.span, modifier.span], **rd_kwargs))
                # elif isinstance(x, MetaDate):
                #     # if hasattr(x, "month"):
                #     #     print("relative", x.month, modifier)
                #     # if hasattr(x, "season"):
                #     #     print("relative", x.season, modifier)
                #     new.append(x)
                elif isinstance(x, MetaRelative):

                    if x.rd.weekday is not None:
                        # eg "next" tuesday adds +1 week
                        weeks = locale.MODIFIERS[modifier.x]
                        # "this" also adds +1, "past" should be -1
                        if weeks == 0:
                            weeks = 1
                        x.rd.weeks += weeks
                        modifier = False

                    new.append(x)

            elif isinstance(x, MetaUnit):
                meta_units.append(x)
                modifier = False
            else:
                if not isinstance(x, MetaOrdinal):
                    new.append(x)
                # meta_units = []
                # modifier = False
        cleaned_bundles.append(new)
    cleaned_bundles = [x for x in cleaned_bundles if any([isinstance(y, MetaRelative) for y in x])]
    return cleaned_bundles


def flatten_inner(l):
    span = []
    for s in l:
        if isinstance(s, list):
            span.extend(flatten_span(s))
        else:
            span.append(s)
    return span


def flatten_span(l):
    return sorted(set(flatten_inner(l)))


def get_min_level(cleaned_bundle):
    level = 100
    for x in cleaned_bundle:
        if isinstance(x, (MetaRelative, MetaUnit)):
            level = min(x.level, level)
    return level


def datify(cleaned_bundle, past_boundary, now):
    span = flatten_span([x.span for x in cleaned_bundle])
    level = get_min_level(cleaned_bundle)
    # print(level)
    rdt, erdt = resolve_dt(cleaned_bundle, now)

    if erdt is not None:
        start_date = rdt + resolve_rd(cleaned_bundle)
        end_date = erdt + resolve_rd(cleaned_bundle)
        start_date, _, _ = resolve_end_period(start_date, level, past_boundary, now)
        end_date = erase_level(erdt, level)
    else:
        resolved_rd = resolve_rd(cleaned_bundle)
        try:
            start_date = rdt + resolved_rd
        except TypeError:
            resolved_rd.years = int(resolved_rd.years)
            resolved_rd.months = int(resolved_rd.months)
            start_date = rdt + resolved_rd

        start_date, end_date, _ = resolve_end_period(start_date, level, past_boundary, now)
    return start_date, end_date, level, span


# flexible for running with a default init of locale_en
en_scanner = Scanner(locale_en)


def has_between(bundle):
    return any(isinstance(x, MetaBetween) for x in bundle)


def merge_dt(mdt, dt):
    # mdt is a made-up mutable dt
    # month is precomputed so that day can be init
    month = mdt[1] if mdt[1] > -1 else dt.month
    day = mdt[2] if mdt[2] > -1 else dt.day
    ndt = datetime(year=mdt[0] if mdt[0] > -1 else dt.year,
                   month=month,
                   day=min(day, BOUNDARIES[month]),
                   hour=mdt[3] if mdt[3] > -1 else dt.hour,
                   minute=mdt[4] if mdt[4] > -1 else dt.minute,
                   second=mdt[5] if mdt[5] > -1 else dt.second,
                   microsecond=mdt[6] if mdt[6] > -1 else dt.microsecond)
    return ndt


def resolve_dt(cleaned_bundle, now):
    # the concept of checking whether there is overlap is good, and crash?
    indices = {'year': 0, 'month': 1, 'day': 2, 'hour': 3,
               'minute': 4, 'second': 5, 'microsecond': 6}
    dts = [-1, -1, -1, -1, -1, -1, -1]
    edts = [-1, -1, -1, -1, -1, -1, -1]
    contains_between = has_between(cleaned_bundle)
    # for d in cleaned_bundle:
    #     if isinstance(d, MetaDate):
    #         for x in d.__dict__:
    #             if x not in ['level', 'span', 'now']:
    #                 if dts[indices[x]] != -1:
    #                     if not contains_between:
    #                         raise ValueError("Field already known in dt, should not overwrite?")
    #                 else:
    #                     dts[indices[x]] = getattr(d, x)
    #                 edts[indices[x]] = getattr(d, x)
    dt = merge_dt(dts, now)
    edt = merge_dt(edts, now) if contains_between else None
    return dt, edt


def resolve_rd(cleaned_bundle):
    rds = rd()
    for d in cleaned_bundle:
        if isinstance(d, MetaRelative):
            rds += d.rd
    return rds


def handle_meta_range(cleaned_bundle, past_boundary, locale, text):
    now = datetime.now()
    phase = 0
    mrange = None
    relatives = []
    metadates = []
    for x in cleaned_bundle:
        if phase == 0 and isinstance(x, MetaRange):
            phase = 1
            mrange = x
        elif phase == 1 or phase == 2 and isinstance(x, MetaRelative):
            phase = 2
            relatives.append(x)
        elif phase == 2 or phase == 3 and isinstance(x, MetaDate):
            phase = 3
            metadates.append(x)
    if phase < 2:
        return None
    # case 1: MetaRange, MetaRelative, MetaRelative, etc
    # log("relatives", relatives, True)
    # log("metadates", metadates, True)
    if not metadates:
        # in this case, the start_date becomes "now" adjusted for level
        # the end_date is the start_date + relativedeltas following
        level = get_min_level(relatives)
        rds = rd()
        start_date = erase_level(now, 1)
        for x in relatives:
            if x.rd is None:
                continue
            rds += x.rd
        try:
            end_date = start_date + rds
        except TypeError:
            rds.years = int(rds.years)
            rds.months = int(rds.months)
            end_date = start_date + rds
        span = flatten_span([mrange.span, [x.span for x in relatives]])
        if start_date > end_date:
            start_date, end_date = end_date, start_date
        return MetaPeriod(start_date, end_date, level, span, locale.NAME, text)
    else:
        spans = flatten_span([flatten_span(x.span) for x in relatives])
        words = [text[x[0]:x[1]].lower() for x in spans]
        # generic
        dt, _ = resolve_dt(metadates, now)
        dt_level = get_min_level(metadates)
        rd_level = get_min_level(relatives)
        if "first" in words:
            # this is actually the logic for !first! days of, not "next days of"
            # case 2: MetaRange, MetaRelative, MetaRelative, etc, MetaDate, Metadate, etc
            start_date, _, date_changed = resolve_end_period(dt, dt_level, past_boundary, now)
            end_date = erase_level(dt, dt_level) + resolve_rd(relatives)
        elif "last" in words:
            _, end_date, date_changed = resolve_end_period(dt, dt_level, past_boundary, now)
            start_date = erase_level(end_date, dt_level) + resolve_rd(relatives)
        else:
            raise NotImplementedError("What's the case here?")
    if date_changed:
        start_date = start_date + rd(years=1)  # I think also here?
        end_date = end_date + rd(years=1)
    span = flatten_span([mrange.span] + [x.span for x in relatives] +
                        [x.span for x in metadates])
    return MetaPeriod(start_date, end_date, min(rd_level, dt_level), span, locale.NAME, text)


def parse_date(text, future=rd(years=30), past=rd(years=100), locale=locale_en, reference_date=None, multi=False, verbose=False,
               return_early=0):
    now = reference_date or datetime.now()
    past = past or rd()
    future = future or rd()
    log("\nSentence", text, verbose)
    scanner = en_scanner if locale == locale_en else Scanner(locale)
    matches, _ = scanner.scan(text)
    log("matches", matches, verbose=verbose)
    parts = get_relevant_parts(matches)
    log("1", parts, verbose=verbose)
    if return_early == 1:
        return parts
    merged = [merge_ordinal_unit(x, text) for x in parts]
    if return_early == 2:
        return merged
    log("2", merged, verbose=verbose)
    cleaned_parts = cleanup_relevant_parts(merged, locale)
    if return_early == 3:
        return cleaned_parts
    log("3", cleaned_parts, verbose)

    if not cleaned_parts:
        return [] if multi else None

    cleaned_parts = sorted(cleaned_parts, key=len, reverse=True)

    mps = []
    past_boundary = now - past
    future_boundary = now + future
    for cleaned_bundle in cleaned_parts:
        if mps and mps[0] and not multi:
            return mps[0]
        handle_meta_result = handle_meta_range(cleaned_bundle, past_boundary, locale, text)
        if handle_meta_result:
            log("4meta", handle_meta_result, verbose)
            mps.append(handle_meta_result)
            continue

        start_date, end_date, level, span = datify(cleaned_bundle, past_boundary, now)

        mp = MetaPeriod(start_date, end_date, level, span, locale.NAME, text)
        log("4default", mp, verbose)
        if past_boundary <= mp.start_date <= future_boundary:
            mps.append(mp)
        else:
            mps.append(None)
    if not multi:
        return mps[0]

    return mps
